from dataclasses import dataclass
from pathlib import Path

import typer
from packaging.requirements import Requirement
from packaging.version import parse as version_parse

from troml.config import SUPPORTED_LIBRARIES


@dataclass
class Classifier:
    """Base classifier."""

    classifiers: set

    def handle(self, *args, **kwargs) -> None:
        raise NotImplementedError()


@dataclass
class TypingTypedClassifier(Classifier):
    """Typing :: Typed classifier."""

    def handle(self, cwd: Path) -> None:
        if (cwd / "py.typed").exists():
            typer.secho(" - Add typed classifier based on py.typed file", fg=typer.colors.GREEN)
            self.classifiers.add("Typing :: Typed")


@dataclass
class LicenseClassifier(Classifier):
    """Deal with the legacy license classifier."""

    def handle(self, project: dict) -> None:
        if project.get("license") or project.get("license-files"):
            for classifier in list(self.classifiers):
                if classifier.startswith("License :: "):
                    typer.secho(
                        " - Remove legacy license classifier (https://peps.python.org/pep-0639/#deprecate-license-classifiers)",
                        fg=typer.colors.RED,
                    )
                    self.classifiers.remove(classifier)


@dataclass
class PythonClassifier(Classifier):
    """Python classifiers."""

    def handle(self, project: dict) -> None:
        if requires_python := project.get("requires-python"):
            typer.secho(f" - Python requirement: {requires_python}", fg=typer.colors.GREEN)

            dependency = f"Python{requires_python}"
            DependencyClassifier(self.classifiers).handle(dependency)


@dataclass
class DependenciesClassifier(Classifier):
    """Dependencies classifiers."""

    def handle(self, data: dict) -> None:
        for dependency in self.get_dependencies(data=data):
            DependencyClassifier(self.classifiers).handle(dependency)

    def get_dependencies(self, data: dict) -> list[str]:
        project = data.get("project", {})
        tool = data.get("tool", {})

        dependencies = []

        if project_dependencies := project.get("dependencies", []):
            for dependency in project_dependencies:
                dependencies.append(dependency)

        for _, group_dependencies in project.get("dependency-groups", {}).items():
            for dependency in group_dependencies:
                # `dependency`` can be a `dict` with a "include-group" key, however those will already be added to the
                # list of dependencies
                if isinstance(dependency, str):
                    dependencies.append(dependency)

        if constraint_dependencies := tool.get("uv", {}).get("constraint-dependencies", []):
            for dependency in constraint_dependencies:
                dependencies.append(dependency)

        # TODO: Handle non-standard tools here, e.g. poetry, etc

        return dependencies


@dataclass
class DependencyClassifier(Classifier):
    """Dependency classifiers"""

    def handle(self, dependency: str) -> None:
        requirement = Requirement(dependency)

        if library := SUPPORTED_LIBRARIES.get(requirement.name):
            specifier_versions = self.get_specifier_versions(requirement=requirement)
            new_classifiers: set = set()

            for version in library.versions:
                potential_classifier = ""

                if not version:
                    potential_classifier = library.classifier_base
                elif requirement.specifier.contains(version) or version in specifier_versions:
                    potential_classifier = f"{library.classifier_base} :: {version}"
                else:
                    continue

                if potential_classifier and potential_classifier not in (self.classifiers | new_classifiers):
                    new_classifiers.add(potential_classifier)

            self.classifiers.update(new_classifiers)

            if len(new_classifiers) > 1:
                typer.secho(f" - Add {len(new_classifiers)} classifiers for {requirement.name}", fg=typer.colors.GREEN)
            elif len(new_classifiers) == 1:
                typer.secho(f" - Add classifier for {requirement.name}", fg=typer.colors.GREEN)

    def get_specifier_versions(self, requirement: Requirement) -> set[str]:
        """Get the major.minor versions in requirement specifier for potential inclusion in classifiers.

        This ensures `Python>=3.9.7` requirement includes "Programming Language :: Python :: 3.9" in classifiers
        """

        specifier_versions = set()

        for specifier in requirement.specifier:
            version = version_parse(specifier.version)
            major_minor_version = f"{version.major}.{version.minor}"

            if major_minor_version not in specifier_versions:
                specifier_versions.add(major_minor_version)

        return specifier_versions



from pathlib import Path
from tomllib import load as toml_load
from typing import Any

import typer


def get_cwd_and_pyproject_data(path: Path | None = None) -> tuple[Path, dict]:
    path = path or Path(".").resolve()

    data = get_pyproject_data(path)
    cwd = path

    if not path.is_dir():
        cwd = path.parent

    return (cwd, data)


def get_pyproject_data(path: Path) -> dict[str, Any]:
    pyproject_path = path

    if path.is_dir():
        pyproject_path = path / "pyproject.toml"

    typer.secho(f"Loading data from '{pyproject_path.parent}'", fg=typer.colors.BLUE)

    with pyproject_path.open("rb") as f:
        return dict(toml_load(f))



import re
from dataclasses import dataclass, field

from trove_classifiers import all_classifiers


@dataclass
class Library:
    name: str
    classifier_pattern: str
    classifier_base: str
    versions: set = field(default_factory=set)

    def __init__(self, name: str, classifier_base: str, classifier_pattern: str):
        self.name = name
        self.classifier_base = classifier_base
        self.classifier_pattern = classifier_pattern
        self.classifier_re = re.compile(classifier_pattern)

        self.set_versions_from_classifiers()

    def set_versions_from_classifiers(self):
        self.versions = set()

        for classifier in all_classifiers:
            if match := self.classifier_re.match(classifier):
                version = match.groupdict().get("version")

                self.versions.add(version)


@dataclass
class Libraries:
    libraries: list[Library] = field(default_factory=list)

    def add(self, library: Library):
        self.libraries.append(library)

    def get(self, name: str) -> Library | None:
        return next(filter(lambda library: library.name.lower() == name.lower(), self.libraries), None)

    def __iter__(self):
        return self.libraries.__iter__()



from collections.abc import Iterable

import typer


def echo_classifiers(title: str, classifiers: Iterable) -> None:
    typer.secho(title, fg=typer.colors.BLUE)

    typer.echo("classifiers = [")

    for classifier in classifiers:
        typer.echo(f'  "{classifier}",')

    typer.echo("]")



from troml.objects import Libraries, Library

DJANGO_LIBRARY = Library(
    name="django",
    classifier_base="Framework :: Django",
    classifier_pattern=r"^Framework :: Django( :: (?P<version>\d+(\.\d+)?))?$",
)
PYTHON_LIBRARY = Library(
    name="python",
    classifier_base="Programming Language :: Python",
    classifier_pattern=r"^Programming Language :: Python( :: (?P<version>\d+(\.\d+)?))?$",
)

SUPPORTED_LIBRARIES = Libraries(
    [
        DJANGO_LIBRARY,
        PYTHON_LIBRARY,
    ]
)



from copy import copy
from pathlib import Path
from typing import Annotated

import typer

from troml.classifiers import DependenciesClassifier, LicenseClassifier, PythonClassifier, TypingTypedClassifier
from troml.loader import get_cwd_and_pyproject_data
from troml.utils import echo_classifiers

app = typer.Typer()


@app.command()
def suggest(path: Annotated[Path, typer.Argument(help="The path of the pyproject.toml.")] = Path(".")) -> None:
    """Suggest new trove classifiers for a project."""

    (cwd, data) = get_cwd_and_pyproject_data(path)

    project = data.get("project", {})
    classifiers = set(project.get("classifiers", []))
    current_classifiers = copy(classifiers)

    TypingTypedClassifier(classifiers).handle(cwd=cwd)
    LicenseClassifier(classifiers).handle(project=project)
    PythonClassifier(classifiers).handle(project=project)
    DependenciesClassifier(classifiers).handle(data=data)

    # Print out current and suggested classifiers
    echo_classifiers("\nCurrent classifiers TOML", current_classifiers)
    echo_classifiers("\nSuggested classifiers TOML", sorted(classifiers))


if __name__ == "__main__":
    app()
